// generated by lv2ttl2c from
// http://gareus.org/oss/lv2/phaserotate#stereo

extern const LV2_Descriptor* lv2_descriptor(uint32_t index);
extern const LV2UI_Descriptor* lv2ui_descriptor(uint32_t index);

static const RtkLv2Description _plugin_stereo = {
	&lv2_descriptor,
	&lv2ui_descriptor
	, 1 // uint32_t dsp_descriptor_id
	, 0 // uint32_t gui_descriptor_id
	, "Phaserotate Stereo" // const char *plugin_human_id
	, (const struct LV2Port[9])
	{
		{ "control", ATOM_IN, nan, nan, nan, "UI to plugin communication"},
		{ "notify", ATOM_OUT, nan, nan, nan, "Plugin to GUI communication"},
		{ "latency", CONTROL_OUT, nan, 0.000000, 8192.000000, "Signal Latency"},
		{ "angleL", CONTROL_IN, 0.000000, -180.000000, 180.000000, "Angle Left"},
		{ "inL", AUDIO_IN, nan, nan, nan, "In Left"},
		{ "outL", AUDIO_OUT, nan, nan, nan, "Out Left"},
		{ "angleR", CONTROL_IN, 0.000000, -180.000000, 180.000000, "Angle Right"},
		{ "inR", AUDIO_IN, nan, nan, nan, "In Right"},
		{ "outR", AUDIO_OUT, nan, nan, nan, "Out Right"},
	}
	, 9 // uint32_t nports_total
	, 2 // uint32_t nports_audio_in
	, 2 // uint32_t nports_audio_out
	, 0 // uint32_t nports_midi_in
	, 0 // uint32_t nports_midi_out
	, 1 // uint32_t nports_atom_in
	, 1 // uint32_t nports_atom_out
	, 3 // uint32_t nports_ctrl
	, 2 // uint32_t nports_ctrl_in
	, 1 // uint32_t nports_ctrl_out
	, 8192 // uint32_t min_atom_bufsiz
	, false // bool send_time_info
	, 2 // uint32_t latency_ctrl_port
};
