// generated by lv2ttl2c from
// http://gareus.org/oss/lv2/darc#stereo

extern const LV2_Descriptor* lv2_descriptor(uint32_t index);
extern const LV2UI_Descriptor* lv2ui_descriptor(uint32_t index);

static const RtkLv2Description _plugin_stereo = {
	&lv2_descriptor,
	&lv2ui_descriptor
	, 1 // uint32_t dsp_descriptor_id
	, 0 // uint32_t gui_descriptor_id
	, "x42-comp - Dynamic Compressor Stereo" // const char *plugin_human_id
	, (const struct LV2Port[14])
	{
		{ "enable", CONTROL_IN, 1.000000, 0.000000, 1.000000, "Enable"},
		{ "hold", CONTROL_IN, 0.000000, 0.000000, 1.000000, "Hold"},
		{ "inputgain", CONTROL_IN, 0.000000, -10.000000, 30.000000, "Input Gain"},
		{ "threshold", CONTROL_IN, -30.000000, -50.000000, -10.000000, "Threshold"},
		{ "Ratio", CONTROL_IN, 0.000000, 0.000000, 1.000000, "Ratio"},
		{ "attack", CONTROL_IN, 0.010000, 0.001000, 0.100000, "Attack Time"},
		{ "release", CONTROL_IN, 0.300000, 0.030000, 3.000000, "Release Time"},
		{ "gain_min", CONTROL_OUT, nan, -20.000000, 40.000000, "Gain Min"},
		{ "gain_max", CONTROL_OUT, nan, -20.000000, 40.000000, "Gain Max"},
		{ "rms", CONTROL_OUT, nan, -80.000000, 10.000000, "Signal Level"},
		{ "inL", AUDIO_IN, nan, nan, nan, "In Left"},
		{ "outL", AUDIO_OUT, nan, nan, nan, "Out Left"},
		{ "inR", AUDIO_IN, nan, nan, nan, "In Right"},
		{ "outR", AUDIO_OUT, nan, nan, nan, "Out Right"},
	}
	, 14 // uint32_t nports_total
	, 2 // uint32_t nports_audio_in
	, 2 // uint32_t nports_audio_out
	, 0 // uint32_t nports_midi_in
	, 0 // uint32_t nports_midi_out
	, 0 // uint32_t nports_atom_in
	, 0 // uint32_t nports_atom_out
	, 10 // uint32_t nports_ctrl
	, 7 // uint32_t nports_ctrl_in
	, 3 // uint32_t nports_ctrl_out
	, 8192 // uint32_t min_atom_bufsiz
	, false // bool send_time_info
	, UINT32_MAX // uint32_t latency_ctrl_port
};
